package com.mediahub

import akka.NotUsed
import akka.actor.ActorSystem
import akka.stream.ActorAttributes.supervisionStrategy
import akka.stream.Supervision.resumingDecider
import akka.stream.alpakka.csv.scaladsl.{CsvParsing, CsvToMap}
import akka.stream.scaladsl.{Compression, FileIO, Flow, Sink, Source}
import akka.util.ByteString
import org.springframework.stereotype.Component

import java.io.File
import java.nio.charset.StandardCharsets
import java.nio.file.Paths
import scala.collection.{immutable, mutable}
import scala.concurrent.Await
import scala.concurrent.duration.DurationInt
import scala.language.postfixOps
import scala.util.Try


@Component
class MovieQueryService extends MovieServiceTrait {

  implicit val system: ActorSystem = ActorSystem()

  val titleBasicsResource: File = new File("src/main/resources/title.basics.tsv.gz")
  val titlePrincipalsResource: File = new File("src/main/resources/title.principals.tsv.gz")
  val nameBasicsResource: File = new File("src/main/resources/name.basics.tsv.gz")
  val titleEpisodeResource: File = new File("src/main/resources/title.episode.tsv.gz")

  def fileSource(file: File): Source[ByteString, NotUsed] = {
    Source.single(file)
      .flatMapConcat(f => FileIO.fromPath(Paths.get(f.getPath), 1 * 1024 * 1024 * 2))
      .via(Compression.gunzip())
  }

  val lineParser: Flow[ByteString, Map[String, String], NotUsed] = {
    CsvParsing
      .lineScanner(CsvParsing.Tab, CsvParsing.DoubleQuote, CsvParsing.DoubleQuote)
      .via(CsvToMap.toMapAsStringsCombineAll(headerPlaceholder = Option.empty))
  }

  override def principalsForMovieName(name: String): List[Principal] = {

    val titleId: String = getIdOfTitle(titleBasicsResource, name)
    val personsIdList: List[String] = getIdOfPersons(titlePrincipalsResource, titleId).toList.flatten
    val personsList = getPersons(nameBasicsResource, personsIdList)
    personsList
  }

  override def tvSeriesWithGreatestNumberOfEpisodes(): List[TvSeries] = {
    val seriesTitleMap = getSeriesTitle(titleEpisodeResource)
    val titleListMaxEpisode = getTitleByIds(titleBasicsResource, seriesTitleMap)
    titleListMaxEpisode
  }



  def getIdOfTitle(file: File, titleName: String): String = {
    val result = fileSource(file)
      .mapAsync(50) {
        res =>
          Source.single(res)
            .via(lineParser)
            .filter(row => row.getOrElse("primaryTitle", "") == titleName)
            .map(a => a.get("tconst"))
            .withAttributes(supervisionStrategy(resumingDecider))
            .runWith(Sink.head)
      }
      .withAttributes(supervisionStrategy(resumingDecider))
      .runWith(Sink.head)

    Await.result(result, 5 minutes)
    result.value.get.get.get
  }

  def getTitleByIds(file: File, titleIdMap: mutable.Map[String, Int]): List[TvSeries] = {
    val result = Source.single(file)
      .flatMapConcat(f => FileIO.fromPath(Paths.get(f.getPath), 1 * 1024 * 1024))
      .via(Compression.gunzip())
      .via(CsvParsing.lineScanner(CsvParsing.Tab, CsvParsing.DoubleQuote, CsvParsing.DoubleQuote))
      .via(CsvToMap.toMapAsStringsCombineAll(StandardCharsets.UTF_8, Option.empty))
      .filter(row => titleIdMap.keySet.toList.contains(row.getOrElse("tconst", "")))
      .map(a => TvSeries(a("originalTitle"), Try(a.getOrElse("startYear","").toInt).getOrElse(0), a.getOrElse("endYear","").toIntOption, a.get("genres").toList))
      .runWith(Sink.collection)

    Await.result(result, 5 minutes)
    result.value.get.get.toList
  }

  def getIdOfPersons(file: File, titleId: String): immutable.Iterable[Option[String]] = {
    val result = fileSource(file)
      .mapAsync(50) {
        res =>
          Source.single(res)
            .via(lineParser)
            .filter(row => row.getOrElse("tconst", "") == titleId)
            .map(a => a.get("nconst"))
            .withAttributes(supervisionStrategy(resumingDecider))
            .runWith(Sink.collection)
      }
      .withAttributes(supervisionStrategy(resumingDecider))
      .runWith(Sink.head)

    Await.result(result, 5 minutes)
  }


  def getPersons(file: File, personsIdList: List[String]): List[Principal] = {
    val result = Source.single(file)
      .flatMapConcat(f => FileIO.fromPath(Paths.get(f.getPath), 1 * 1024 * 1024))
      .via(Compression.gunzip())
      .via(CsvParsing.lineScanner(CsvParsing.Tab, CsvParsing.DoubleQuote, CsvParsing.DoubleQuote))
      .via(CsvToMap.toMapAsStringsCombineAll(StandardCharsets.UTF_8, Option.empty))
      .filter(row => personsIdList.contains(row.getOrElse("nconst", "")))
      .map(a => Principal(a("primaryName"), Try(a("birthYear").toInt).getOrElse(0), a("deathYear").toIntOption, a.get("primaryProfession").toList))
      .runWith(Sink.collection)

    Await.result(result, 5 minutes)
    result.value.get.get.toList
  }

  def getSeriesTitle(file: File): mutable.Map[String, Int] = {

    val seriesMap = collection.mutable.Map("0" -> 0)
    val maxSeriesMap = collection.mutable.Map("0" -> 0)

    val result = Source.single(file)
      .flatMapConcat(f => FileIO.fromPath(Paths.get(f.getPath), 1 * 1024 * 1024))
      .via(Compression.gunzip())
      .via(CsvParsing.lineScanner(CsvParsing.Tab, CsvParsing.DoubleQuote, CsvParsing.DoubleQuote))
      .via(CsvToMap.toMapAsStringsCombineAll(StandardCharsets.UTF_8, Option.empty))
      .map(row => seriesMap += seriesMap.get(row("parentTconst")).map(x => row("parentTconst") -> (x + 1)).getOrElse(row("parentTconst") -> 1))
      .runWith(Sink.seq)

    Await.result(result, 5 minutes)

    for (_ <- 1 to 10) {
      maxSeriesMap.addOne(seriesMap.maxBy(x => x._2)._1 -> seriesMap.maxBy(x => x._2)._2)
      seriesMap.remove(seriesMap.maxBy(x => x._2)._1)
    }
    maxSeriesMap
  }

}